@@ -7,9 +7,10 @@ import (
 	"sync/atomic"
 	"testing"
 
-	"k8s.io/api/core/v1"
+	v1 "k8s.io/api/core/v1"
 	knet "k8s.io/api/networking/v1"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
+	"k8s.io/apimachinery/pkg/labels"
 	"k8s.io/apimachinery/pkg/runtime"
 	"k8s.io/apimachinery/pkg/types"
 	"k8s.io/apimachinery/pkg/watch"
@@ -17,6 +18,16 @@ import (
 	core "k8s.io/client-go/testing"
 	"k8s.io/client-go/tools/cache"
 
+	egressfirewall "github.com/ovn-org/ovn-kubernetes/go-controller/pkg/crd/egressfirewall/v1"
+	egressfirewallfake "github.com/ovn-org/ovn-kubernetes/go-controller/pkg/crd/egressfirewall/v1/apis/clientset/versioned/fake"
+
+	apiextensions "k8s.io/apiextensions-apiserver/pkg/apis/apiextensions/v1beta1"
+	apiextensionsfake "k8s.io/apiextensions-apiserver/pkg/client/clientset/clientset/fake"
+
+	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/config"
+	egressip "github.com/ovn-org/ovn-kubernetes/go-controller/pkg/crd/egressip/v1"
+	egressipfake "github.com/ovn-org/ovn-kubernetes/go-controller/pkg/crd/egressip/v1/apis/clientset/versioned/fake"
+
 	. "github.com/onsi/ginkgo"
 	. "github.com/onsi/gomega"
 )
@@ -98,6 +109,46 @@ func newService(name, namespace string) *v1.Service {
 	}
 }
 
+func newEgressFirewall(name, namespace string) *egressfirewall.EgressFirewall {
+	return &egressfirewall.EgressFirewall{
+		ObjectMeta: newObjectMeta(name, namespace),
+		Spec: egressfirewall.EgressFirewallSpec{
+			[]egressfirewall.EgressFirewallRule{
+				{
+					Type: egressfirewall.EgressFirewallRuleAllow,
+					To: egressfirewall.EgressFirewallDestination{
+						CIDRSelector: "1.2.3.4/32",
+					},
+				},
+			},
+		},
+	}
+}
+
+func newEgressIP(name, namespace string) *egressip.EgressIP {
+	return &egressip.EgressIP{
+		ObjectMeta: newObjectMeta(name, namespace),
+		Spec: egressip.EgressIPSpec{
+			EgressIPs: []string{
+				"192.168.126.10",
+			},
+		},
+	}
+
+}
+
+func newCRD(name, namespace string) *apiextensions.CustomResourceDefinition {
+	return &apiextensions.CustomResourceDefinition{
+		ObjectMeta: metav1.ObjectMeta{
+			Name: name,
+			UID:  types.UID(name),
+			Labels: map[string]string{
+				"name": name,
+			},
+		},
+	}
+}
+
 func objSetup(c *fake.Clientset, objType string, listFn func(core.Action) (bool, runtime.Object, error)) *watch.FakeWatcher {
 	w := watch.NewFake()
 	c.AddWatchReactor(objType, core.DefaultWatchReactor(w, nil))
@@ -105,6 +156,27 @@ func objSetup(c *fake.Clientset, objType string, listFn func(core.Action) (bool,
 	return w
 }
 
+func egressFirewallObjSetup(c *egressfirewallfake.Clientset, objType string, listFn func(core.Action) (bool, runtime.Object, error)) *watch.FakeWatcher {
+	w := watch.NewFake()
+	c.AddWatchReactor(objType, core.DefaultWatchReactor(w, nil))
+	c.AddReactor("list", objType, listFn)
+	return w
+}
+
+func crdObjSetup(c *apiextensionsfake.Clientset, objType string, listFn func(core.Action) (bool, runtime.Object, error)) *watch.FakeWatcher {
+	w := watch.NewFake()
+	c.AddWatchReactor(objType, core.DefaultWatchReactor(w, nil))
+	c.AddReactor("list", objType, listFn)
+	return w
+}
+
+func egressIPObjSetup(c *egressipfake.Clientset, objType string, listFn func(core.Action) (bool, runtime.Object, error)) *watch.FakeWatcher {
+	w := watch.NewFake()
+	c.AddWatchReactor(objType, core.DefaultWatchReactor(w, nil))
+	c.AddReactor("list", objType, listFn)
+	return w
+}
+
 type handlerCalls struct {
 	added   int32
 	updated int32
@@ -126,20 +198,35 @@ func (c *handlerCalls) getDeleted() int {
 var _ = Describe("Watch Factory Operations", func() {
 	var (
 		fakeClient                                *fake.Clientset
+		egressIPFakeClient                        *egressipfake.Clientset
+		egressFirewallFakeClient                  *egressfirewallfake.Clientset
+		crdFakeClient                             *apiextensionsfake.Clientset
 		podWatch, namespaceWatch, nodeWatch       *watch.FakeWatcher
 		policyWatch, endpointsWatch, serviceWatch *watch.FakeWatcher
+		egressFirewallWatch, crdWatch             *watch.FakeWatcher
+		egressIPWatch                             *watch.FakeWatcher
 		pods                                      []*v1.Pod
 		namespaces                                []*v1.Namespace
 		nodes                                     []*v1.Node
 		policies                                  []*knet.NetworkPolicy
 		endpoints                                 []*v1.Endpoints
 		services                                  []*v1.Service
+		egressIPs                                 []*egressip.EgressIP
 		wf                                        *WatchFactory
+		egressFirewalls                           []*egressfirewall.EgressFirewall
+		crds                                      []*apiextensions.CustomResourceDefinition
 		err                                       error
 	)
 
 	BeforeEach(func() {
+
+		// Restore global default values before each testcase
+		config.PrepareTestConfig()
+
 		fakeClient = &fake.Clientset{}
+		egressFirewallFakeClient = &egressfirewallfake.Clientset{}
+		crdFakeClient = &apiextensionsfake.Clientset{}
+		egressIPFakeClient = &egressipfake.Clientset{}
 
 		pods = make([]*v1.Pod, 0)
 		podWatch = objSetup(fakeClient, "pods", func(core.Action) (bool, runtime.Object, error) {
@@ -194,23 +281,54 @@ var _ = Describe("Watch Factory Operations", func() {
 			}
 			return true, obj, nil
 		})
+
+		egressFirewalls = make([]*egressfirewall.EgressFirewall, 0)
+		egressFirewallWatch = egressFirewallObjSetup(egressFirewallFakeClient, "egressfirewalls", func(core.Action) (bool, runtime.Object, error) {
+			obj := &egressfirewall.EgressFirewallList{}
+			for _, p := range egressFirewalls {
+				obj.Items = append(obj.Items, *p)
+			}
+			return true, obj, nil
+		})
+
+		crds = make([]*apiextensions.CustomResourceDefinition, 0)
+		crdWatch = crdObjSetup(crdFakeClient, "customresourcedefinitions", func(core.Action) (bool, runtime.Object, error) {
+			obj := &apiextensions.CustomResourceDefinitionList{}
+			for _, p := range crds {
+				obj.Items = append(obj.Items, *p)
+			}
+			return true, obj, nil
+		})
+
+		egressIPs = make([]*egressip.EgressIP, 0)
+		egressIPWatch = egressIPObjSetup(egressIPFakeClient, "egressips", func(core.Action) (bool, runtime.Object, error) {
+			obj := &egressip.EgressIPList{}
+			for _, p := range egressIPs {
+				obj.Items = append(obj.Items, *p)
+			}
+			return true, obj, nil
+		})
 	})
 
 	AfterEach(func() {
 		wf.Shutdown()
+		if wf.efFactory != nil {
+			wf.ShutdownEgressFirewallWatchFactory()
+		}
 	})
 
 	Context("when a processExisting is given", func() {
-		testExisting := func(objType reflect.Type, namespace string, lsel *metav1.LabelSelector) {
-			wf, err = NewWatchFactory(fakeClient)
+		testExisting := func(objType reflect.Type, namespace string, sel labels.Selector) {
+			wf, err = NewWatchFactory(fakeClient, egressIPFakeClient, egressFirewallFakeClient, crdFakeClient)
 			Expect(err).NotTo(HaveOccurred())
-			h, err := wf.addHandler(objType, namespace, lsel,
+			err = wf.InitializeEgressFirewallWatchFactory()
+			Expect(err).NotTo(HaveOccurred())
+			h := wf.addHandler(objType, namespace, sel,
 				cache.ResourceEventHandlerFuncs{},
 				func(objs []interface{}) {
 					defer GinkgoRecover()
 					Expect(len(objs)).To(Equal(1))
 				})
-			Expect(err).NotTo(HaveOccurred())
 			Expect(h).NotTo(BeNil())
 			wf.removeHandler(objType, h)
 		}
@@ -245,22 +363,43 @@ var _ = Describe("Watch Factory Operations", func() {
 			testExisting(serviceType, "", nil)
 		})
 
+		It("is called for each existing egressFirewall", func() {
+			egressFirewalls = append(egressFirewalls, newEgressFirewall("myEgressFirewall", "default"))
+			testExisting(egressFirewallType, "", nil)
+		})
+		It("is called for each existing CRDS", func() {
+			crds = append(crds, newCRD("myCRD", ""))
+			testExisting(crdType, "", nil)
+		})
+		It("is called for each existing egressIP", func() {
+			egressIPs = append(egressIPs, newEgressIP("myEgressIP", "default"))
+			testExisting(egressIPType, "", nil)
+		})
+
 		It("is called for each existing pod that matches a given namespace and label", func() {
 			pod := newPod("pod1", "default")
 			pod.ObjectMeta.Labels["blah"] = "foobar"
 			pods = append(pods, pod)
-			testExisting(podType, "default", &metav1.LabelSelector{
-				MatchLabels: map[string]string{"blah": "foobar"},
-			})
+
+			sel, err := metav1.LabelSelectorAsSelector(
+				&metav1.LabelSelector{
+					MatchLabels: map[string]string{"blah": "foobar"},
+				},
+			)
+			Expect(err).NotTo(HaveOccurred())
+
+			testExisting(podType, "default", sel)
 		})
 	})
 
 	Context("when existing items are known to the informer", func() {
 		testExisting := func(objType reflect.Type) {
-			wf, err = NewWatchFactory(fakeClient)
+			wf, err = NewWatchFactory(fakeClient, egressIPFakeClient, egressFirewallFakeClient, crdFakeClient)
+			Expect(err).NotTo(HaveOccurred())
+			err = wf.InitializeEgressFirewallWatchFactory()
 			Expect(err).NotTo(HaveOccurred())
 			var addCalls int32
-			h, err := wf.addHandler(objType, "", nil,
+			h := wf.addHandler(objType, "", nil,
 				cache.ResourceEventHandlerFuncs{
 					AddFunc: func(obj interface{}) {
 						atomic.AddInt32(&addCalls, 1)
@@ -268,7 +407,6 @@ var _ = Describe("Watch Factory Operations", func() {
 					UpdateFunc: func(old, new interface{}) {},
 					DeleteFunc: func(obj interface{}) {},
 				}, nil)
-			Expect(err).NotTo(HaveOccurred())
 			Expect(int(addCalls)).To(Equal(2))
 			wf.removeHandler(objType, h)
 		}
@@ -308,11 +446,38 @@ var _ = Describe("Watch Factory Operations", func() {
 			services = append(services, newService("myservice2", "default"))
 			testExisting(serviceType)
 		})
+		It("calls ADD for each existing egressFirewall", func() {
+			egressFirewalls = append(egressFirewalls, newEgressFirewall("myFirewall", "default"))
+			egressFirewalls = append(egressFirewalls, newEgressFirewall("myFirewall1", "default"))
+			testExisting(egressFirewallType)
+		})
+		It("calls ADD for each existing CRD", func() {
+			crds = append(crds, newCRD("crd1", ""))
+			crds = append(crds, newCRD("crd2", ""))
+			testExisting(crdType)
+		})
+		It("calls ADD for each existing egressIP", func() {
+			egressIPs = append(egressIPs, newEgressIP("myEgressIP", "default"))
+			egressIPs = append(egressIPs, newEgressIP("myEgressIP1", "default"))
+			testExisting(egressIPType)
+		})
 	})
 
-	addFilteredHandler := func(wf *WatchFactory, objType reflect.Type, namespace string, lsel *metav1.LabelSelector, funcs cache.ResourceEventHandlerFuncs) (*Handler, *handlerCalls) {
+	Context("when EgressIP is disabled", func() {
+		testExisting := func(objType reflect.Type) {
+			wf, err = NewWatchFactory(fakeClient, egressIPFakeClient, egressFirewallFakeClient, crdFakeClient)
+			Expect(err).NotTo(HaveOccurred())
+			Expect(wf.informers).NotTo(HaveKey(objType))
+		}
+		It("does not contain Egress IP informer", func() {
+			config.OVNKubernetesFeature.EnableEgressIP = false
+			testExisting(egressIPType)
+		})
+	})
+
+	addFilteredHandler := func(wf *WatchFactory, objType reflect.Type, namespace string, sel labels.Selector, funcs cache.ResourceEventHandlerFuncs) (*Handler, *handlerCalls) {
 		calls := handlerCalls{}
-		h, err := wf.addHandler(objType, namespace, lsel, cache.ResourceEventHandlerFuncs{
+		h := wf.addHandler(objType, namespace, sel, cache.ResourceEventHandlerFuncs{
 			AddFunc: func(obj interface{}) {
 				defer GinkgoRecover()
 				atomic.AddInt32(&calls.added, 1)
@@ -329,7 +494,6 @@ var _ = Describe("Watch Factory Operations", func() {
 				funcs.DeleteFunc(obj)
 			},
 		}, nil)
-		Expect(err).NotTo(HaveOccurred())
 		Expect(h).NotTo(BeNil())
 		return h, &calls
 	}
@@ -339,7 +503,7 @@ var _ = Describe("Watch Factory Operations", func() {
 	}
 
 	It("responds to pod add/update/delete events", func() {
-		wf, err = NewWatchFactory(fakeClient)
+		wf, err = NewWatchFactory(fakeClient, egressIPFakeClient, egressFirewallFakeClient, crdFakeClient)
 		Expect(err).NotTo(HaveOccurred())
 
 		added := newPod("pod1", "default")
@@ -373,11 +537,12 @@ var _ = Describe("Watch Factory Operations", func() {
 	})
 
 	It("responds to multiple pod add/update/delete events", func() {
-		wf, err = NewWatchFactory(fakeClient)
+		wf, err = NewWatchFactory(fakeClient, egressIPFakeClient, egressFirewallFakeClient, crdFakeClient)
 		Expect(err).NotTo(HaveOccurred())
 
 		const nodeName string = "mynode"
 		type opTest struct {
+			mu      sync.Mutex
 			pod     *v1.Pod
 			added   int
 			updated int
@@ -396,6 +561,8 @@ var _ = Describe("Watch Factory Operations", func() {
 				pod := obj.(*v1.Pod)
 				ot, ok := testPods[pod.Name]
 				Expect(ok).To(BeTrue())
+				ot.mu.Lock()
+				defer ot.mu.Unlock()
 				Expect(ot.added).To(BeNumerically("<", 2))
 				ot.added++
 			},
@@ -403,6 +570,8 @@ var _ = Describe("Watch Factory Operations", func() {
 				newPod := new.(*v1.Pod)
 				ot, ok := testPods[newPod.Name]
 				Expect(ok).To(BeTrue())
+				ot.mu.Lock()
+				defer ot.mu.Unlock()
 				Expect(ot.updated).To(BeNumerically("<", 2))
 				ot.updated++
 				Expect(newPod.Spec.NodeName).To(Equal(nodeName))
@@ -411,6 +580,8 @@ var _ = Describe("Watch Factory Operations", func() {
 				pod := obj.(*v1.Pod)
 				ot, ok := testPods[pod.Name]
 				Expect(ok).To(BeTrue())
+				ot.mu.Lock()
+				defer ot.mu.Unlock()
 				Expect(ot.deleted).To(BeNumerically("<", 2))
 				ot.deleted++
 			},
@@ -421,7 +592,9 @@ var _ = Describe("Watch Factory Operations", func() {
 			for _, ot := range testPods {
 				pods = append(pods, ot.pod)
 				podWatch.Add(ot.pod)
+				ot.mu.Lock()
 				ot.pod.Spec.NodeName = nodeName
+				ot.mu.Unlock()
 				podWatch.Modify(ot.pod)
 				pods = pods[:0]
 				podWatch.Delete(ot.pod)
@@ -434,16 +607,18 @@ var _ = Describe("Watch Factory Operations", func() {
 		Eventually(c.getUpdated, 2).Should(Equal(10))
 		Eventually(c.getDeleted, 2).Should(Equal(10))
 		for _, ot := range testPods {
+			ot.mu.Lock()
 			Expect(ot.added).Should(Equal(2))
 			Expect(ot.updated).Should(Equal(2))
 			Expect(ot.deleted).Should(Equal(2))
+			ot.mu.Unlock()
 		}
 
 		wf.RemovePodHandler(h)
 	})
 
 	It("responds to namespace add/update/delete events", func() {
-		wf, err = NewWatchFactory(fakeClient)
+		wf, err = NewWatchFactory(fakeClient, egressIPFakeClient, egressFirewallFakeClient, crdFakeClient)
 		Expect(err).NotTo(HaveOccurred())
 
 		added := newNamespace("default")
@@ -477,7 +652,7 @@ var _ = Describe("Watch Factory Operations", func() {
 	})
 
 	It("responds to node add/update/delete events", func() {
-		wf, err = NewWatchFactory(fakeClient)
+		wf, err = NewWatchFactory(fakeClient, egressIPFakeClient, egressFirewallFakeClient, crdFakeClient)
 		Expect(err).NotTo(HaveOccurred())
 
 		added := newNode("mynode")
@@ -511,10 +686,11 @@ var _ = Describe("Watch Factory Operations", func() {
 	})
 
 	It("responds to multiple node add/update/delete events", func() {
-		wf, err = NewWatchFactory(fakeClient)
+		wf, err = NewWatchFactory(fakeClient, egressIPFakeClient, egressFirewallFakeClient, crdFakeClient)
 		Expect(err).NotTo(HaveOccurred())
 
 		type opTest struct {
+			mu      sync.Mutex
 			node    *v1.Node
 			added   int
 			updated int
@@ -533,6 +709,8 @@ var _ = Describe("Watch Factory Operations", func() {
 				node := obj.(*v1.Node)
 				ot, ok := testNodes[node.Name]
 				Expect(ok).To(BeTrue())
+				ot.mu.Lock()
+				defer ot.mu.Unlock()
 				Expect(ot.added).To(BeNumerically("<", 2))
 				ot.added++
 			},
@@ -540,6 +718,8 @@ var _ = Describe("Watch Factory Operations", func() {
 				newNode := new.(*v1.Node)
 				ot, ok := testNodes[newNode.Name]
 				Expect(ok).To(BeTrue())
+				ot.mu.Lock()
+				defer ot.mu.Unlock()
 				Expect(ot.updated).To(BeNumerically("<", 2))
 				ot.updated++
 				Expect(newNode.Status.Phase).To(Equal(v1.NodeTerminated))
@@ -548,8 +728,10 @@ var _ = Describe("Watch Factory Operations", func() {
 				node := obj.(*v1.Node)
 				ot, ok := testNodes[node.Name]
 				Expect(ok).To(BeTrue())
+				ot.mu.Lock()
 				Expect(ot.deleted).To(BeNumerically("<", 2))
 				ot.deleted++
+				ot.mu.Unlock()
 			},
 		})
 
@@ -558,7 +740,9 @@ var _ = Describe("Watch Factory Operations", func() {
 			for _, ot := range testNodes {
 				nodes = append(nodes, ot.node)
 				nodeWatch.Add(ot.node)
+				ot.mu.Lock()
 				ot.node.Status.Phase = v1.NodeTerminated
+				ot.mu.Unlock()
 				nodeWatch.Modify(ot.node)
 				nodes = nodes[:0]
 				nodeWatch.Delete(ot.node)
@@ -571,9 +755,11 @@ var _ = Describe("Watch Factory Operations", func() {
 		Eventually(c.getUpdated, 2).Should(Equal(10))
 		Eventually(c.getDeleted, 2).Should(Equal(10))
 		for _, ot := range testNodes {
+			ot.mu.Lock()
 			Expect(ot.added).Should(Equal(2))
 			Expect(ot.updated).Should(Equal(2))
 			Expect(ot.deleted).Should(Equal(2))
+			ot.mu.Unlock()
 		}
 
 		wf.RemoveNodeHandler(h)
@@ -581,6 +767,7 @@ var _ = Describe("Watch Factory Operations", func() {
 
 	It("correctly orders queued informer initial add events and subsequent update events", func() {
 		type opTest struct {
+			mu      sync.Mutex
 			node    *v1.Node
 			added   int
 			updated int
@@ -595,7 +782,7 @@ var _ = Describe("Watch Factory Operations", func() {
 			nodes = append(nodes, node)
 		}
 
-		wf, err = NewWatchFactory(fakeClient)
+		wf, err = NewWatchFactory(fakeClient, egressIPFakeClient, egressFirewallFakeClient, crdFakeClient)
 		Expect(err).NotTo(HaveOccurred())
 
 		startWg := sync.WaitGroup{}
@@ -619,6 +806,8 @@ var _ = Describe("Watch Factory Operations", func() {
 				node := obj.(*v1.Node)
 				ot, ok := testNodes[node.Name]
 				Expect(ok).To(BeTrue())
+				ot.mu.Lock()
+				defer ot.mu.Unlock()
 				Expect(ot.added).To(Equal(0), "add for node %s already run", node.Name)
 				ot.added++
 			},
@@ -628,6 +817,8 @@ var _ = Describe("Watch Factory Operations", func() {
 				ot, ok := testNodes[newNode.Name]
 				Expect(ok).To(BeTrue())
 				// Expect updates to be processed after Add
+				ot.mu.Lock()
+				defer ot.mu.Unlock()
 				Expect(ot.added).To(Equal(1), "update for node %s processed before initial add!", newNode.Name)
 				Expect(ot.updated).To(Equal(0))
 				ot.updated++
@@ -639,14 +830,18 @@ var _ = Describe("Watch Factory Operations", func() {
 
 		// Adds are done synchronously at handler addition time
 		for _, ot := range testNodes {
+			ot.mu.Lock()
 			Expect(ot.added).To(Equal(1), "missing add for node %s", ot.node.Name)
+			ot.mu.Unlock()
 		}
 		Expect(c.getAdded()).To(Equal(len(testNodes)))
 
 		// Updates are async and may take a bit longer to finish
 		Eventually(c.getUpdated, 10).Should(Equal(len(testNodes)))
 		for _, ot := range testNodes {
+			ot.mu.Lock()
 			Expect(ot.updated).To(Equal(1), "missing update for node %s", ot.node.Name)
+			ot.mu.Unlock()
 		}
 
 		wf.RemoveNodeHandler(h)
@@ -654,6 +849,7 @@ var _ = Describe("Watch Factory Operations", func() {
 
 	It("correctly orders serialized informer initial add events and subsequent update events", func() {
 		type opTest struct {
+			mu        sync.Mutex
 			namespace *v1.Namespace
 			added     int
 			updated   int
@@ -668,7 +864,7 @@ var _ = Describe("Watch Factory Operations", func() {
 			namespaces = append(namespaces, namespace)
 		}
 
-		wf, err = NewWatchFactory(fakeClient)
+		wf, err = NewWatchFactory(fakeClient, egressIPFakeClient, egressFirewallFakeClient, crdFakeClient)
 		Expect(err).NotTo(HaveOccurred())
 
 		startWg := sync.WaitGroup{}
@@ -692,6 +888,8 @@ var _ = Describe("Watch Factory Operations", func() {
 				namespace := obj.(*v1.Namespace)
 				ot, ok := testNamespaces[namespace.Name]
 				Expect(ok).To(BeTrue())
+				ot.mu.Lock()
+				defer ot.mu.Unlock()
 				Expect(ot.added).To(Equal(0))
 				ot.added++
 			},
@@ -701,6 +899,8 @@ var _ = Describe("Watch Factory Operations", func() {
 				ot, ok := testNamespaces[newNamespace.Name]
 				Expect(ok).To(BeTrue())
 				// Expect updates to be processed after Add
+				ot.mu.Lock()
+				defer ot.mu.Unlock()
 				Expect(ot.added).To(Equal(1), "update for namespace %s processed before initial add!", newNamespace.Name)
 				Expect(ot.updated).To(Equal(0))
 				ot.updated++
@@ -712,21 +912,25 @@ var _ = Describe("Watch Factory Operations", func() {
 
 		// Adds are done synchronously at handler addition time
 		for _, ot := range testNamespaces {
+			ot.mu.Lock()
 			Expect(ot.added).To(Equal(1), "missing add for namespace %s", ot.namespace.Name)
+			ot.mu.Unlock()
 		}
 		Expect(c.getAdded()).To(Equal(len(testNamespaces)))
 
 		// Updates are async and may take a bit longer to finish
 		Eventually(c.getUpdated, 10).Should(Equal(len(testNamespaces)))
 		for _, ot := range testNamespaces {
+			ot.mu.Lock()
 			Expect(ot.updated).To(Equal(1), "missing update for namespace %s", ot.namespace.Name)
+			ot.mu.Unlock()
 		}
 
 		wf.RemoveNamespaceHandler(h)
 	})
 
 	It("responds to policy add/update/delete events", func() {
-		wf, err = NewWatchFactory(fakeClient)
+		wf, err = NewWatchFactory(fakeClient, egressIPFakeClient, egressFirewallFakeClient, crdFakeClient)
 		Expect(err).NotTo(HaveOccurred())
 
 		added := newPolicy("mypolicy", "default")
@@ -760,7 +964,7 @@ var _ = Describe("Watch Factory Operations", func() {
 	})
 
 	It("responds to endpoints add/update/delete events", func() {
-		wf, err = NewWatchFactory(fakeClient)
+		wf, err = NewWatchFactory(fakeClient, egressIPFakeClient, egressFirewallFakeClient, crdFakeClient)
 		Expect(err).NotTo(HaveOccurred())
 
 		added := newEndpoints("myendpoints", "default")
@@ -801,7 +1005,7 @@ var _ = Describe("Watch Factory Operations", func() {
 	})
 
 	It("responds to service add/update/delete events", func() {
-		wf, err = NewWatchFactory(fakeClient)
+		wf, err = NewWatchFactory(fakeClient, egressIPFakeClient, egressFirewallFakeClient, crdFakeClient)
 		Expect(err).NotTo(HaveOccurred())
 
 		added := newService("myservice", "default")
@@ -834,8 +1038,109 @@ var _ = Describe("Watch Factory Operations", func() {
 		wf.RemoveServiceHandler(h)
 	})
 
+	It("responds to egressFirewall add/update/delete events", func() {
+		wf, err = NewWatchFactory(fakeClient, egressIPFakeClient, egressFirewallFakeClient, crdFakeClient)
+		err = wf.InitializeEgressFirewallWatchFactory()
+		Expect(err).NotTo(HaveOccurred())
+
+		added := newEgressFirewall("myEgressFirewall", "default")
+		h, c := addHandler(wf, egressFirewallType, cache.ResourceEventHandlerFuncs{
+			AddFunc: func(obj interface{}) {
+				egressFirewall := obj.(*egressfirewall.EgressFirewall)
+				Expect(reflect.DeepEqual(egressFirewall, added)).To(BeTrue())
+			},
+			UpdateFunc: func(old, new interface{}) {
+				newEgressFirewall := new.(*egressfirewall.EgressFirewall)
+				Expect(reflect.DeepEqual(newEgressFirewall, added)).To(BeTrue())
+				Expect(newEgressFirewall.Spec.Egress[0].Type).To(Equal(egressfirewall.EgressFirewallRuleDeny))
+			},
+			DeleteFunc: func(obj interface{}) {
+				egressFirewall := obj.(*egressfirewall.EgressFirewall)
+				Expect(reflect.DeepEqual(egressFirewall, added)).To(BeTrue())
+			},
+		})
+
+		egressFirewalls = append(egressFirewalls, added)
+		egressFirewallWatch.Add(added)
+		Eventually(c.getAdded, 2).Should(Equal(1))
+		added.Spec.Egress[0].Type = egressfirewall.EgressFirewallRuleDeny
+		egressFirewallWatch.Modify(added)
+		Eventually(c.getUpdated, 2).Should(Equal(1))
+		egressFirewalls = egressFirewalls[:0]
+		egressFirewallWatch.Delete(added)
+		Eventually(c.getDeleted, 2).Should(Equal(1))
+
+		wf.RemoveEgressFirewallHandler(h)
+	})
+	It("responds to crd add/update/delete events", func() {
+		wf, err = NewWatchFactory(fakeClient, egressIPFakeClient, egressFirewallFakeClient, crdFakeClient)
+		Expect(err).NotTo(HaveOccurred())
+
+		added := newCRD("crd1", "")
+		h, c := addHandler(wf, crdType, cache.ResourceEventHandlerFuncs{
+			AddFunc: func(obj interface{}) {
+				crd := obj.(*apiextensions.CustomResourceDefinition)
+				Expect(reflect.DeepEqual(crd, added)).To(BeTrue())
+			},
+			UpdateFunc: func(old, new interface{}) {
+				newcrd := new.(*apiextensions.CustomResourceDefinition)
+				Expect(reflect.DeepEqual(newcrd, added)).To(BeTrue())
+				Expect(newcrd.Spec.Group).To(Equal("my-test"))
+			},
+			DeleteFunc: func(obj interface{}) {
+				crd := obj.(*apiextensions.CustomResourceDefinition)
+				Expect(reflect.DeepEqual(crd, added)).To(BeTrue())
+			},
+		})
+
+		crds = append(crds, added)
+		crdWatch.Add(added)
+		Eventually(c.getAdded, 2).Should(Equal(1))
+		added.Spec.Group = "my-test"
+		crdWatch.Modify(added)
+		Eventually(c.getUpdated, 2).Should(Equal(1))
+		crds = crds[:0]
+		crdWatch.Delete(added)
+		Eventually(c.getDeleted, 2).Should(Equal(1))
+
+		wf.RemoveCRDHandler(h)
+
+	})
+	It("responds to egressIP add/update/delete events", func() {
+		wf, err = NewWatchFactory(fakeClient, egressIPFakeClient, egressFirewallFakeClient, crdFakeClient)
+		Expect(err).NotTo(HaveOccurred())
+
+		added := newEgressIP("myEgressIP", "default")
+		h, c := addHandler(wf, egressIPType, cache.ResourceEventHandlerFuncs{
+			AddFunc: func(obj interface{}) {
+				egressIP := obj.(*egressip.EgressIP)
+				Expect(reflect.DeepEqual(egressIP, added)).To(BeTrue())
+			},
+			UpdateFunc: func(old, new interface{}) {
+				newEgressIP := new.(*egressip.EgressIP)
+				Expect(reflect.DeepEqual(newEgressIP, added)).To(BeTrue())
+				Expect(newEgressIP.Spec.EgressIPs).To(Equal([]string{"192.168.126.10"}))
+			},
+			DeleteFunc: func(obj interface{}) {
+				egressIP := obj.(*egressip.EgressIP)
+				Expect(reflect.DeepEqual(egressIP, added)).To(BeTrue())
+			},
+		})
+
+		egressIPs = append(egressIPs, added)
+		egressIPWatch.Add(added)
+		Eventually(c.getAdded, 2).Should(Equal(1))
+		added.Spec.EgressIPs = []string{"192.168.126.10"}
+		egressIPWatch.Modify(added)
+		Eventually(c.getUpdated, 2).Should(Equal(1))
+		egressIPs = egressIPs[:0]
+		egressIPWatch.Delete(added)
+		Eventually(c.getDeleted, 2).Should(Equal(1))
+
+		wf.RemoveEgressIPHandler(h)
+	})
 	It("stops processing events after the handler is removed", func() {
-		wf, err = NewWatchFactory(fakeClient)
+		wf, err = NewWatchFactory(fakeClient, egressIPFakeClient, egressFirewallFakeClient, crdFakeClient)
 		Expect(err).NotTo(HaveOccurred())
 
 		added := newNamespace("default")
@@ -864,7 +1169,7 @@ var _ = Describe("Watch Factory Operations", func() {
 	})
 
 	It("filters correctly by label and namespace", func() {
-		wf, err = NewWatchFactory(fakeClient)
+		wf, err = NewWatchFactory(fakeClient, egressIPFakeClient, egressFirewallFakeClient, crdFakeClient)
 		Expect(err).NotTo(HaveOccurred())
 
 		passesFilter := newPod("pod1", "default")
@@ -874,12 +1179,17 @@ var _ = Describe("Watch Factory Operations", func() {
 		failsFilter2 := newPod("pod3", "otherns")
 		failsFilter2.ObjectMeta.Labels["blah"] = "foobar"
 
-		_, c := addFilteredHandler(wf,
-			podType,
-			"default",
+		sel, err := metav1.LabelSelectorAsSelector(
 			&metav1.LabelSelector{
 				MatchLabels: map[string]string{"blah": "foobar"},
 			},
+		)
+		Expect(err).NotTo(HaveOccurred())
+
+		_, c := addFilteredHandler(wf,
+			podType,
+			"default",
+			sel,
 			cache.ResourceEventHandlerFuncs{
 				AddFunc: func(obj interface{}) {
 					pod := obj.(*v1.Pod)
@@ -928,19 +1238,24 @@ var _ = Describe("Watch Factory Operations", func() {
 	})
 
 	It("correctly handles object updates that cause filter changes", func() {
-		wf, err = NewWatchFactory(fakeClient)
+		wf, err = NewWatchFactory(fakeClient, egressIPFakeClient, egressFirewallFakeClient, crdFakeClient)
 		Expect(err).NotTo(HaveOccurred())
 
 		pod := newPod("pod1", "default")
 		pod.ObjectMeta.Labels["blah"] = "baz"
 
+		sel, err := metav1.LabelSelectorAsSelector(
+			&metav1.LabelSelector{
+				MatchLabels: map[string]string{"blah": "foobar"},
+			},
+		)
+		Expect(err).NotTo(HaveOccurred())
+
 		equalPod := pod
 		h, c := addFilteredHandler(wf,
 			podType,
 			"default",
-			&metav1.LabelSelector{
-				MatchLabels: map[string]string{"blah": "foobar"},
-			},
+			sel,
 			cache.ResourceEventHandlerFuncs{
 				AddFunc: func(obj interface{}) {
 					p := obj.(*v1.Pod)
